// *************************************************************************
//    BindEnv.h
// *************************************************************************

#ifndef _BindEnv_H_
#define _BindEnv_H_

#include <iostream.h>

#include "HashVal.h"
#include "Term.h"
#include "Entity.h"
#include "Debug.h"

// *************************************************************************
// A bindenv is a set of substitutions, mapping variables to Terms.
// Variables in a Clause are numbered sequentially,
// with each variable's number stored in the 'var_num' field of a CArg.
// A binding can be implemented as an array, mapping
// 'var_num' numbers to Terms.
//
// **  WARNING:
//  IF      a bindenv is allocated within a function and deleted on 
//          function exit, 
//  THEN  1) you MUST create a StackMark first.
//        2) before returning from the function you MUST pop back to 
//          the stackmark.  Otherwise the trail may contain 
//          references to the defunct bindenv, and cause 
//          a crash later.
// ***
// *************************************************************************

class Cor_Term;
class Cor_CArg;

// *************************************************************************

class Cor_BindEnv {

  private:
    int _limit;

    Cor_Term *bindings;     // Array of variable bindings

    Cor_BindEnv();

  public:

#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif

    Cor_BindEnv(int size);
   ~Cor_BindEnv();    

    inline Cor_Term lookup(Variable var) const {
        ASSERT((var >= 0) && (var < _limit));
        return(bindings[var]);
    }
   
    inline Cor_CArg *carg_lookup(Variable var) {
        ASSERT((var >= 0) && (var < _limit));
        ASSERT(bindings[var].bindenv == NULL);
        return(&(bindings[var].expr));
    }

    inline Cor_CArg *full_lookup(Variable var, Cor_BindEnv **var_env) {
        ASSERT((var >= 0) && (var < _limit));
        *var_env = bindings[var].bindenv;
        return(&(bindings[var].expr));
    }
    
    inline int not_bound(Variable var) const {
        ASSERT((var >= 0) && (var < _limit));
        return(bindings[var].expr.is_empty());
    }
    
    inline int try_lookup(Cor_Term &term, Variable var) const {
        ASSERT((var >= 0) && (var < _limit));
        if (bindings[var].expr.is_empty())
            return(0);
        term = bindings[var];
        return(1);
    }
    
    inline HashVal hash_var(Variable var) const {
        ASSERT((var >= 0) && (var < _limit));
        Cor_Term *term = &(bindings[var]);
        if (term->expr.is_empty())
            return(VarHashValue);
        return(term->expr.hash(term->bindenv));
    }
    
    inline int bind(Variable var, const Cor_Term& binding) {
        ASSERT((var >= 0) && (var < _limit));
        bindings[var] = binding;
        return(1);
    }

    inline int bind(Variable var, const Cor_CArg *gnd_arg) {
        ASSERT((var >= 0) && (var < _limit));
        bindings[var].expr    = *gnd_arg;
        bindings[var].bindenv = NULL;
        return(1);
    }

    inline void clear(Variable var) {
        ASSERT((var >= 0) && (var < _limit));
        bindings[var].clear();
    }

      // Clear all bindings
    void clear_all() {
        for (int j=0; j < _limit; j++)
            clear(j);
    }

      // Avoid using this (slow)
    Cor_BindEnv *dup();

#ifdef    NOT_USED
    Cor_BindEnv *copy_shell();
    void restore(Cor_BindEnv *env);
#endif

    void resize(int newsize);    // Aborts
    int all_free();

    inline int max_var_number() { return(_limit); }

    void printon(ostream *out);
    friend ostream& operator <<(ostream& out, const Cor_BindEnv& env);
};

#include "Term.i"

// *************************************************************************

#endif _BindEnv_H_
