
//**************************************************************************
//    Connection.C
//
//    The Connection Class - Used to hold all information about a 
//    connection. This includes the socket being used, a ParseStack, and
//    an array of users.
//
//**************************************************************************

#include <iostream.h>
#include <iomanip.h>

#include "Connection.h"
#include "Database.h"
#include "Globals.h"
#include "Debug.h"

#include "Suspension.h"
#include "CoralStream.h"

//**************************************************************************
// Connection Constructor -
//    Clears out the user array, and puts the client's stream onto the
//    parse stack. This function also generates a default user "nobody"
//    so the client has an execution envirornment before they login(). 
//
//**************************************************************************
Cor_Connection::Cor_Connection(CoralStream *io_strm) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    num_users = 0;
    nxt_user = 1;           // Since 0 is "nobody"
    cur_user = NOBODY;      // "nobody" is the current user

    for (int i = 0; i < MAX_USRS_PER_CONN; i++)
        userArray[i] = (Cor_User *) NULL;

    if (io_strm)
        conn_strm = io_strm;
    else
        conn_strm = cin_strm;

      // Generate user "nobody" (id is always 0)
    if (io_strm)
        userArray[NOBODY] = new Cor_User(io_strm,io_strm,io_strm);
    else
        userArray[NOBODY] = new Cor_User(cin_strm, cout_strm, cerr_strm);

      // Put the connection onto the parse stack
    if (! io_strm) {
        pStack.push(cin_strm);
    } else {
        pStack.push(io_strm);
    }

    strcpy((userArray[NOBODY])->login_name,"nobody");

      // No path to start with
    cwd[0] = '\0';
    sock_fd = -1;
}

//**************************************************************************
// Connection Destructor -
//    Deletes all users in the user array, and deletes the default
//    user "nobody".
//
//**************************************************************************
Cor_Connection::~Cor_Connection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i=1; i < MAX_USRS_PER_CONN; i++)
        if (userArray[i])
            DeleteUser(i);

      // Now get rid of nobody
    Cor_suspension->killall(conn_id, 0);
    delete userArray[0];

      // Close the connection
    if (conn_strm && (conn_strm != cin_strm))
        delete conn_strm;
}

//**************************************************************************
// NewUser() -
//    Generates a new user using the name and password given, and sets
//    them up. If this new user is the only user on this connection,
//    (excluding "nobody"), the current user is switched to them. 
//
//    Return: The user_id of the newly generated user, or < 0 if an
//        error occurs.
//
//**************************************************************************
int Cor_Connection::NewUser(char *name, char *passwd)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!name || !(*name))
        return(-1);

      // user "nobody" is not counted in num_users
    if (num_users >= (MAX_USRS_PER_CONN-1))
        return(-1);

      // Do not allow user nobody to be created in this fashion
    if (strcmp(name,"nobody") == 0)
        return(-1);

      // NYI - Check that this is this users correct password


    while(userArray[nxt_user])
        nxt_user = (nxt_user + 1) % MAX_USRS_PER_CONN;

    ASSERT(conn_strm);
    CoralStream *io  = new CoralStream(conn_strm->rdbuf());

    userArray[nxt_user] = new Cor_User(io,io,io);
    strncpy((userArray[nxt_user])->login_name, name, 9);
    strncpy((userArray[nxt_user])->password, passwd, 9);

    (userArray[nxt_user])->user_id  = nxt_user;

      // This is added simply as a convienence, if you are nobody
      // (like when you first connect) login() also switches you
      // to be the new user you just created.
    if (cur_user == NOBODY)
        useUser(nxt_user);

    num_users++;
    return(nxt_user);
}

//**************************************************************************
// DeleteUser() -
//    Deletes the user specified by user_id, freeing up the space in the
//    user array and deleting any suspended queries they have.
//    If they are the current user, the current user becomes the default
//    user "nobody".  The user "nobody" cannot be deleted with this
//    function, they are only deleted when the connection is closed.
//
//**************************************************************************
void Cor_Connection::DeleteUser(int user_id)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ((! user_id) || (! userArray[user_id]))
        return;

      // delete suspended queries for this user
    Cor_suspension->killall(conn_id, user_id);

    delete userArray[user_id];
    userArray[user_id] = (Cor_User *) NULL;

    nxt_user = user_id;

    if (user_id == cur_user){
      // Logging self out - become "nobody"
      useUser(NOBODY);
    }

    num_users--;
}

//**************************************************************************
// useConnection() -
//    Switch all global pointers to make this connection the current
//    connection. This function also switches the global current user
//    to be the current user of this connection. 
//
//    Return: 1 on success, 0 if current user of this connection didn't
//        exist.
//
//**************************************************************************
int Cor_Connection::useConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_curConn = this;

      // Set the global parse stack to this one
    Cor_parseStack = &pStack;
    if (! Cor_parseStack->is_empty() )
        Cor_parseStack->reset_position();

    if (!useUser(cur_user)){
        // The user specified doesn't exist - become "nobody"
        useUser(NOBODY);
        return(0);
    }
    
    return(1);
}

//**************************************************************************
// useUser() -
//    Switch all global pointers to make this user the current
//    user.
//
//    Return: 1 on success, 0 if the user didn't exist.
//
//**************************************************************************
int Cor_Connection::useUser(int new_user)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! userArray[new_user])
        return(0);

    Cor_curUser = userArray[new_user];
    cur_user = new_user;

      // What else needs to be switched ??
    Cor_exEnv = &(Cor_curUser->Cor_exEnv);
    Cor_CurDB = Cor_curUser->Cor_CurDB;

    return(1);
}

//**************************************************************************
// FindUser() -
//    Finds the user given by name. If the name is not unique, returns
//    the first user that matches.
//
//    Return: The user_id of the user with this name, or < 0 if no user
//        by that name is on this connection. 
//
//**************************************************************************
int Cor_Connection::FindUser(char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i=0; i < MAX_USRS_PER_CONN; i++)
        if (userArray[i] && (strcmp(name,userArray[i]->login_name) == 0))
            return(i);

    return(-1);
}

//**************************************************************************
// ListWho()
//    Lists all users for this connection, printing the information to
//    the file pointer specified. The default user "nobody" is ignored.
//
//    Return: The number of users found.
//
//**************************************************************************
int Cor_Connection::ListWho(CoralStream *strm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int cnt=0;

    *strm << "Connection " << conn_id << ":\n";
    for (int i=1; i < MAX_USRS_PER_CONN; i++)
        if (userArray[i]){
            cnt++;
            *strm << setw(4) << i
                  << "   " << userArray[i]->login_name << endl;
        }

    return(cnt);
}

//**************************************************************************
