// *************************************************************************
//    PipelinedExecInfo.C
// *************************************************************************

#include <string.h>

#include "PipelinedExecInfo.h"

#include "QFPipelinedModuleData.h"
#include "TupleIterator.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************

#define NULL_INT_VAL -2

// *************************************************************************
Cor_PipelinedExecInfo::Cor_PipelinedExecInfo(Cor_QFPipelinedModuleData *md,
            int pred_offset)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel;
    int           j, offset;

    // Get the first rule defining this relation
    if (!(curRule = (md->pred_ptrs[pred_offset])->clauses))
        return;

    // Extract its RuleInfo 
    rInfo = (md->ruleArray)->rInfoPtrs[curRule->rule_number];

    // Table is the same as that in the QFPipelinedModuleData
    F = md->F;
    indices = md->indices[curRule->rule_number];
    int numVars = md->maxNumVars[pred_offset];

    env = new Cor_BindEnv(numVars);

    int numLits = md->maxNumLiterals[pred_offset];
    tuple_its = new Cor_TupleIterator* [numLits];
    memset((char *)tuple_its, 0, numLits*sizeof(Cor_TupleIterator *) );
    tuples = new Cor_Tuple* [numLits];
    stack_marks = new Cor_StackMark [numLits];

    // Change to use naive backtracking?
    if (md->module_info->NaiveBacktrack)
        md->backtrackInfo.set_naive_backtracking(md->ruleArray);

    firstFailureBacktrack = md->backtrackInfo.
        firstFailureBacktrackArrays[curRule->rule_number];
    nextFailureBacktrack = md->backtrackInfo.
        nextFailureBacktrackArrays[curRule->rule_number];
    successBacktrack = md->backtrackInfo.
        successBacktrackPoints[curRule->rule_number];

    
    for (j = 0; j < rInfo->num_literals; j++) {

        //Check if predicate should be negated, signalled by an offset <= -2.
        if (NEGATED(j))
            offset = (-2) - rInfo->relations[j];
        else
            offset = rInfo->relations[j];
    
        rel = F->rels[offset].relPtr;
        tuple_its[j] = new Cor_TupleIterator(rel, rInfo->arg_list[j],
            env, NULL, NULL, NULL, &(indices[j]));
    }

    current = backtrack_index = 0;

    direction = FORWARD;
    cut_crossed = 0;
}

// *************************************************************************
int Cor_PipelinedExecInfo::init()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(0);
}

// *************************************************************************
Cor_PipelinedExecInfo::~Cor_PipelinedExecInfo()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!curRule)
        return;

    for (int i=0; i< rInfo->num_literals; i++) {
        if (tuple_its[i]) {
            tuple_its[i]->release();
            delete tuple_its[i];
        }
    }

    delete env;
    env  = NULL;

    delete [] tuple_its;
    delete [] tuples;
    delete [] stack_marks;
    indices = NULL;
    curRule = NULL;
    rInfo = NULL;

    F = NULL;
}

// *************************************************************************
//  Go to the next rule that defines this predicate. 
//  Sets up the info in PipelinedExecInfo
//
//  Returns 1 on success, 0 on failure.
// *************************************************************************
int Cor_PipelinedExecInfo::NextRule(Cor_QFPipelinedModuleData *md)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel;
    int           j, offset;

    // if we are already at the last rule defining the predicate
    if (!curRule->pnext) {
        return(0);
    }

    // use the structures allocated earlier
    // after resetting their values
    // The trail has been used already to reset values in the bindenvs
    for (int i=0; i< rInfo->num_literals; i++) {
        if (tuple_its[i]) {
           // tuple_its[i]->release();
            delete tuple_its[i];
        }
        // There might be a memory LEAK here
        tuple_its[i] = NULL;
        tuples[i] = NULL;
    }
    
    curRule = curRule->pnext;

    // Extract its RuleInfo 
    rInfo = (md->ruleArray)->rInfoPtrs[curRule->rule_number];
    indices = md->indices[curRule->rule_number];

    // Initialize all other fields
    firstFailureBacktrack = md->backtrackInfo.
            firstFailureBacktrackArrays[curRule->rule_number];
    nextFailureBacktrack = md->backtrackInfo.
            nextFailureBacktrackArrays[curRule->rule_number];
    successBacktrack = md->backtrackInfo.
            successBacktrackPoints[curRule->rule_number];

    for (j = 0; j < rInfo->num_literals; j++) {
        // Check if predicate should be negated, signaled by an offset <= -2.
        if (NEGATED(j))
            offset = (-2) - rInfo->relations[j];
        else
            offset = rInfo->relations[j];
    
        rel = F->rels[offset].relPtr ;
        tuple_its[j] = new Cor_TupleIterator(rel, rInfo->arg_list[j],
            env, NULL, NULL, NULL, &(indices[j]));
    }

    offset = current = backtrack_index = 0;
    direction = FORWARD;

    return(1);
}

// *************************************************************************
//  Every time evaluate is called a number of structures need to be
//  initialized. However, this is a big overhead if there is going
//  to be allocation and deallocation for each evaluation of a seminaive
//  rule.  So we save them instead of deallocating them, and reuse them
//  on successive calls.  We assume no parallelism however.
//  TO DO:  Since space is never deallocated, we could be in trouble with
//  recursive module calls.  We may need to keep a count of saved up
//  structs, and free them if there are too many saved up structures.

// *************************************************************************

#define MARKLIST(X) ((markOffsets[X]<0) ? NULL : \
            rm_array->markList[markOffsets[X]])

#define NEGATED1(index) (rInfo_relations[index] < 0)


// *************************************************************************
//  Evaluates the query in a top-down fashion. The Table structure
//  F is defined in interp.h. It is a structure containing pointers to the
//  various relations that will have to be accessed. It is
//  initialized at query execution time with pointers to current relations
//  and is a class variable of PipelinedExecInfo.The Solve function
//  accesses all relations via the level of indirection provided by Table.
//
//  The processing for a rule is done in a left-to-right manner
//  The index 'current' counts how far we are along in the join.
//  It represents the position of the current rhs literal being
//  processed.
//
//  We use top-down evaluation with backtracking
// 
//  Returns 1 on success and 0 on failure
// *************************************************************************
int Cor_PipelinedExecInfo::Solve()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // for efficiency
    int rInfo_num_literals = rInfo->num_literals;
    int *rInfo_relations = rInfo->relations;
    Cor_TupleIterator *tuple_its_index;

    // Execution of a Pipelined module - top down using backtracking

    for(; current >= 0;) {

        // Exit statement from for loop. When current < 0, it implies that
        // no more facts can be generated from this rule evaluation, i.e.
        // it has backtracked past its starting point.

        // Check to see if rule evaluation has been succesful.
        // This happens when a valid tuple for the last rhs literal
        // has been found. When this happens, we are going to return

        if ( current >= rInfo_num_literals ) {
            // decrement current so that the next time we begin here 
            current = rInfo_num_literals - 1;
            backtrack_index = successBacktrack;
            direction = BACKWARD;
            return(1);
        }

        tuple_its_index = tuple_its[current];

        // To do a scan of each relation on the rhs, a TupleIterator is
        // needed on that relation. If this is the first time the relation
        // is reached during the evaluation, a TupleIterator needs to be created
        // for it.

        if (direction) {        // == FORWARD
            // iterator needs to be initialized
            tuple_its_index->reset();
            tuple_its_index->calling_pei = this;
            backtrack_index = current;
        }

        if (current == backtrack_index) {
            // Get the next tuple for the current relation in the nested
            // loops join.

            tuple_its[current]->get_next_tuple();

            // If there is a valid tuple, go on to the next rhs relation.
            if ( (!tuple_its_index->no_match() && !NEGATED1(current))
                 || (tuple_its_index->no_match() && NEGATED1(current)) )
            {
                 current++;
                 direction = FORWARD;
                 continue;
            } else {
                backtrack_index = ((direction)
                    ? firstFailureBacktrack[current]
                    : nextFailureBacktrack[current]);
            }
        }

        // On failure to find a valid tuple, backtrack to the previous
        // rhs literal.

        // Undo bindings BEFORE
        // deleting tuple_iterator (and the tuple_env in it).

        tuple_its_index->release(); // releases state of tuple iterator

        // Decrement current so that the join evaluation backtracks to the
        // previous rhs relation, and a fresh tuple for that relation will
        // be fetched in the next iteration through the for loop.
        current--;
        direction = BACKWARD;
    }

    // Have failed 
    stackmark0.pop_to();

    return(0);
}

// *************************************************************************
