
//**************************************************************************
//    PvmStream.C
//
//    The Cor_pvmbuf Class - A buffer class for providing stream access
//    to PVM communication. 
//    (PVM - Parallel Virtual Machine)
//    
//    A Cor_pvmbuf is a streambuf with 2 PVM communication endpoints whose
//    task ids are my_tid and rec_tid. Reading from the Cor_pvmbuf is
//    equivalent to receiving messages from the rec_tid. Writing to the
//    Cor_pvmbuf is equivalent to sending messages to the rec_tid. 
//
//**************************************************************************

#ifdef    DO_PVM

#include "PvmStream.h"


//**************************************************************************
//    Constructor for a client --> server Cor_pvmbuf -
//    The client's tid is obtained by calling pvm_mytid().
//    The server's tid ( = rec_tid) is obtained from the file
//    PVM_COMM_FILE which stores all the services' tids. 
//    
//**************************************************************************

Cor_pvmbuf::Cor_pvmbuf()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    init();

    if ((my_tid = pvm_mytid()) < 0)  {
        pvm_perror("pvm_mytid");
        return;
    }

    // get the server's tid from PVM_COMM_FILE
    rec_tid = -1;
    ifstream server_file(PVM_COMM_FILE);
    if (server_file) {
        char name[100];
        server_file >> name;
        while (!server_file.eof()) {
            if (!strcmp(name, "server")) { // get a server
                if (server_file >> rec_tid) break;
                else {
                    cerr << "error: wrong format in file " << PVM_COMM_FILE << endl;
                    return;
                }
            }
            else {
                cerr << "error: wrong format in file " << PVM_COMM_FILE << endl;
                return;
            }
        }
        server_file.close();
    }
        
    if (rec_tid == -1) {
        cerr << "error: no server running yet\n";
        return;
    }

#ifdef DEBUG_PVM
    cout << "my_tid = " << my_tid << endl;
    cout << "rec_tid = " << rec_tid << endl;
#endif
    
}

//**************************************************************************
//    Constructor for a Cor_pvmbuf with receiver's tid supplied -
//    The tid of the sender is obtained by calling pvm_mytid().
//    The receiver's tid is obtained from the argument.
//    
//**************************************************************************

Cor_pvmbuf::Cor_pvmbuf(int _rec_tid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    init();

    if ((my_tid = pvm_mytid()) < 0)  {
        pvm_perror("pvm_mytid");
        return;
    }

    if ((rec_tid = _rec_tid ) < 0)  {
        cerr << "error: negative tid\n";
        return;
    }

#ifdef DEBUG_PVM
    cout << "my_tid = " << my_tid << endl;
    cout << "rec_tid = " << rec_tid << endl;
#endif
    
}

//**************************************************************************
//    Constructor for a Cor_pvmbuf with both ends' tids supplied -
//    
//**************************************************************************

Cor_pvmbuf::Cor_pvmbuf(int _my_tid, int _rec_tid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    init();

    if (((my_tid = _my_tid) < 0) || ((rec_tid = _rec_tid ) < 0))  {
        cerr << "error: negative tid\n";
        return;
    }

#ifdef DEBUG_PVM
    cout << "my_tid = " << my_tid << endl;
    cout << "rec_tid = " << rec_tid << endl;
#endif
    
}

//**************************************************************************
//    Destructor for a Cor_pvmbuf -
//    The put area is synchronized. 
//    
//**************************************************************************

Cor_pvmbuf::~Cor_pvmbuf()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sync();
}

//**************************************************************************
//    Initialize the Cor_pvmbuf -
//    The sizes are set and the buffer is allocated.
//
//**************************************************************************

void Cor_pvmbuf::init()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    get_buffer_size = put_buffer_size = 4096;
    allocate();
}

//**************************************************************************
//    Check if the Cor_pvmbuf is valid -
//    Return 1 if valid. Return 0 if invalid.
//
//**************************************************************************

int Cor_pvmbuf::Valid()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (my_tid >= 0 && rec_tid >= 0);
}

//**************************************************************************
//    Close the Cor_pvmbuf -
//    The process leaves pvm.
//    It's not called in the destructor since one server can have multiple
//    clients and there are more than one server --> client Cor_pvmbuf.
//    If the close() is called on any one of these Cor_pvmbuf's, then
//    other Cor_pvmbuf's get disconnected.
//    Normally it's called only for a client --> server Cor_pvmbuf.
//    Return 0 if success. Return -1 if pvm error occurs. 
//**************************************************************************

int Cor_pvmbuf::close()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (pvm_exit() < 0) {
        pvm_perror("pvm_exit");
        return -1;
    }
    return 0;
}

//**************************************************************************
//    Allocate the put and get areas -
//    Return 0 if success. Return EOF otherwise.
//
//**************************************************************************

int Cor_pvmbuf::doallocate()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int buffer_size = get_buffer_size + put_buffer_size;
    char *cp = new char[buffer_size];
    if( cp == NULL ) {
        return(EOF);
    }
    setb(cp, cp + buffer_size, 1);
    setg(cp, cp, cp - 1);
    setp(cp + get_buffer_size, cp + buffer_size);
    return 0;
}

//**************************************************************************
//    Synchronize the put area -
//    Return EOF is faliure occurs. 
//    
//**************************************************************************

int Cor_pvmbuf::sync()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return overflow(EOF);
}

//**************************************************************************
//    Consume Characters in the put area -
//    The characters in the put area is sent to receiver (with rec_tid)
//    Return EOF is faliure occurs.
//
//**************************************************************************

int Cor_pvmbuf::overflow(int c)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char * cp = pbase();
    streamsize len = pptr() - cp;

#ifdef DEBUG_PVM
    cout << "OVERFLOW ";
    cout << "  LEN = " << len << endl;
#endif
    
    if (len > 0) {
        pvm_initsend(PvmDataDefault);
        // send the length of the msg first
        if (pvm_pkint(&len, 1, 1) < 0) {
            pvm_perror("pvm_pkint");
            return EOF;
        }

        // send the real msg
        if (pvm_pkbyte((char*)cp, len, 1) < 0) {
            pvm_perror("pvm_pkbyte");
            return EOF;
        }
        if (pvm_send(rec_tid, CORAL_TAG) < 0)  {  
            pvm_perror("pvm_send");
            return EOF;
        }
    }

    setp(pbase(), epptr());
    if( c != EOF ) {
        return sputc(c);
    }
    return 0;
}

//**************************************************************************
//    Supply characters for fetching -
//    Message from receiver (with rec_tid) is received and put to put area. 
//    Return EOF is faliure occurs.
//**************************************************************************

int Cor_pvmbuf::underflow()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *cp = gptr();
    if( egptr() > cp ) {        // chars still in buffer
        // gbump(1);  // Needed?
        return(*cp);
    }
    sync();                     // make sure our output is flushed
                                // before we wait for input
    cp = eback();

    int bufid = pvm_recv(rec_tid, CORAL_TAG);
    if (bufid < 0) {
        pvm_perror("pvm_recv");
        return EOF;
    }
    int len;

    // before every pvm msg there is an integer indicating the length
    // of the msg
    if (pvm_upkint(&len, 1, 1) < 0) {
        pvm_perror("pvm_upkint");
        return EOF;
    }

#ifdef DEBUG_PVM
    cout << "UNDERFLOW ";
    cout << "  LEN = " << len << endl;
#endif
    
    if (pvm_upkbyte(cp, len, 1) < 0) { // assume len <= put_buffer_size
        pvm_perror("pvm_upkbyte");
        return EOF;
    }

    setg(cp, cp, cp + len); // no putback (at least for now...)

    return(*cp);
}

//**************************************************************************

#endif    // DO_PVM
