// *************************************************************************
//    StrRefTable.h - Reference Counting for strings
//
//    Shaun Flisakowski - May 1995
// 
//    This hashtable handles the reference counting for strings, it 
//    deletes them when the references drop to zero.
//
//    It also deletes all strings remaining in the table when it is
//    deleted. If DEBUG is set, it will also complain about those items
//    remaining in the table.
//
// *************************************************************************

#ifndef _STRREF_TABLE_H_
#define _STRREF_TABLE_H_

#include  "HashVal.h"

// *************************************************************************

    // What is a reasonable size for this? (it can grow)
#define    COR_DEFLT_STRREF_TAB_SIZE    4096

// *************************************************************************

class Cor_StrRefTable;

class Cor_StrRefElement {

    friend class Cor_StrRefTable;
    friend class Cor_StrRefCursor;

  private:
        // For linking in buckets
    Cor_StrRefElement *_next;

    void set_next(Cor_StrRefElement *next) { _next = next; }
    Cor_StrRefElement *get_next() { return(_next); }

    char         *_str;
    HashVal       _hash;
    int           _ref_cnt;

    Cor_StrRefElement();
    Cor_StrRefElement(char *str, int init_cnt=0);
   ~Cor_StrRefElement();

      // Complain about the reference count for this string
    void complain();
};

// *************************************************************************

class Cor_StrRefCursor {

    friend class Cor_StrRefTable;

  private:

    int index;                      // The element of the array currently
                                    // being examined
    Cor_StrRefElement    *current;  // NULL indicates to move to next index

    Cor_StrRefTable      *table;    // The symbol table that is
                                    // being scanned 
  public:

    Cor_StrRefCursor(Cor_StrRefTable *table);
   ~Cor_StrRefCursor();

        // To move to the next item (False when done)
    int ScanNext();

        // To get the data out
    char     *str() { return(current->_str); }
    int       cnt() { return(current->_ref_cnt); }
};

// *************************************************************************

class Cor_StrRefTable {

    friend class Cor_StrRefCursor;

private:
    int     _size;                // The size of the table
    int     _num_elements;        // The number of elements currently stored

    int     _element_limit;       // Expand table if the number of elements
                                  // grows larger than this (Currently
                                  //  expanded via doubling size
                                  //  and rehashing all elements)

    Cor_StrRefElement **_reftab;   // The table itself

  public:

        // size of zero uses COR_DEFLT_STRREF_TAB_SIZE
    Cor_StrRefTable(int size = 0);
   ~Cor_StrRefTable();

    int size() { return(_size); }
    int num_elements() { return(_num_elements); }

        // This checks for duplicates before inserting,
        // and will return a pointer to the duplicate,
    char *insert(char *str, int **refer_pntr, int alloc_on_insert=1,
            HashVal *hash=NULL, int hash_valid=0);

        // Does the reference count thing for this string, deleting
        // it if the count goes to zero.
    void destroy(char *str, HashVal *hash=NULL);

        // Returns NULL if string is not in table,
        // otherwise the element containing it.
    Cor_StrRefElement *lookup(char *str, int *indx=NULL, HashVal *hash=NULL,
            int hash_valid=0);

        // Delete all items in the table
    void clear(int _complain=0);

        // Grow the hash table larger (double it)
    void rehash_bigger();

        // For scanning the entire table
    Cor_StrRefCursor *OpenScan();
};

// *************************************************************************

#endif _STRREF_TABLE_H_
